@@ -12,23 +12,32 @@
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 #include <cerrno>
-#include <climits>
 #include <csignal>
 #include <cstdint>
 #include <cstring>
 #include <map>
+#include <memory>
+#include <mutex>
+#include <stdlib.h>
+#include <string>
+#include <sys/select.h>
 #include <t3key/key.h>
-#include <thread>
-#include <transcript/transcript.h>
-
 #include <t3widget/internal.h>
 #include <t3widget/key.h>
 #include <t3widget/keybuffer.h>
 #include <t3widget/log.h>
 #include <t3widget/main.h>
 #include <t3widget/util.h>
+#include <thread>
+#include <transcript/transcript.h>
+#include <type_traits>
+#include <unistd.h>
+#include <utility>
 
-namespace t3_widget {
+#include "t3key/key_errors.h"
+#include "t3window/terminal.h"
+
+namespace t3widget {
 
 #define MAX_SEQUENCE 100
 
@@ -36,6 +45,7 @@ enum {
   WINCH_SIGNAL,
   QUIT_SIGNAL,
   EXIT_MAIN_LOOP_SIGNAL,
+  RESTART_READ_SIGNAL,
 };
 
 struct kp_mapping_t {
@@ -309,6 +319,9 @@ static void read_keys() {
           // \r\n as line endings this will double the number of newlines. As this is the same when
           // not using bracketed paste, this is a acceptable strategy.
           key_buffer.push_back(c == '\n' || c == '\r' ? EKEY_NL : EKEY_PROTECT | c);
+        } else if (c == EKEY_PASTE_START) {
+          in_bracketed_paste = true;
+          key_buffer.push_back(EKEY_PASTE_START);
         } else {
           key_buffer.push_back(c);
         }
@@ -370,8 +383,7 @@ static key_t decode_sequence(bool outer) {
             unget_key_sequence(sequence);
             return -1;
           }
-          return decode_xterm_mouse_sgr_urxvt(sequence.data(), sequence.size()) ? EKEY_MOUSE_EVENT
-                                                                                : -1;
+          return decode_xterm_mouse_sgr_urxvt(sequence) ? EKEY_MOUSE_EVENT : -1;
         } else if (c == '~') {
           if (sequence.size() != 6 || sequence[2] != '2' || sequence[3] != '0') {
             return -1;
@@ -379,12 +391,11 @@ static key_t decode_sequence(bool outer) {
           if (!outer) {
             /* If this is not the outer decode_sequence call, push everything
                back onto the character list, and do nothing. A next call to
-               decode_sequence will take care of the mouse handling. */
+               decode_sequence will take care of the paste handling. */
             unget_key_sequence(sequence);
             return -1;
           }
           if (sequence[4] == '0') {
-            in_bracketed_paste = true;
             return EKEY_PASTE_START;
           }
           return -1;
@@ -521,7 +532,7 @@ static bool is_function_key(const char *str) {
   } while (false)
 
 /* Initialize the key map */
-complex_error_t init_keys(const char *term, bool separate_keypad) {
+complex_error_t init_keys(const optional<std::string> &term, bool separate_keypad) {
   complex_error_t result;
   struct sigaction sa;
   sigset_t sigs;
@@ -537,7 +548,7 @@ complex_error_t init_keys(const char *term, bool separate_keypad) {
     RETURN_ERROR(complex_error_t::SRC_TRANSCRIPT, transcript_error);
   }
 
-  keymap.reset(t3_key_load_map(term, nullptr, &error));
+  keymap.reset(t3_key_load_map(term.is_valid() ? term.value().c_str() : nullptr, nullptr, &error));
   if (keymap == nullptr) {
     RETURN_ERROR(complex_error_t::SRC_T3_KEY, error);
   }
@@ -717,6 +728,11 @@ void reinit_keys() {
   // Enable bracketed paste.
   t3_term_putp("\033[?2004h");
   reinit_mouse_reporting();
+
+  // Notify the read_keys thread that it should restart the select action. This will reinitialize
+  // the file descriptor set, which is necessary for GPM.
+  char restart_read_signal = RESTART_READ_SIGNAL;
+  nosig_write(signal_pipe[1], &restart_read_signal, 1);
 }
 
 void cleanup_keys() {
@@ -776,4 +792,4 @@ void async_safe_exit_main_loop(int exit_code) {
   errno = saved_errno;
 }
 
-};  // namespace
+}  // namespace t3widget
