@@ -11,8 +11,10 @@
    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
+#include <algorithm>
 #include <chrono>
 #include <condition_variable>
+#include <cstdint>
 #include <cstdlib>
 #include <cstring>
 #include <list>
@@ -20,25 +22,28 @@
 #include <mutex>
 #include <string>
 #include <thread>
+#include <utility>
+
+#include "util.h"
+#include "widget_api.h"
 
 #ifdef HAS_SELECT_H
 #include <sys/select.h>
 #else
-#include <sys/types.h>
 #include <unistd.h>
 #endif
 
+#include <sys/select.h>
 #include <t3widget/extclipboard.h>
 #include <t3widget/log.h>
-#include <t3widget/ptr.h>
-
 // FIXME: remove incr_sends on long periods of inactivity
 /* This file used to contain both the Xlib and the XCB implementations for connecting to the X11
    clipboard. However, the Xlib version has been removed, but the code is still structured to have
    multiple implementations. This should be cleaned up at some point. */
 #include <xcb/xcb.h>
+#include <xcb/xproto.h>
 
-namespace t3_widget {
+namespace t3widget {
 
 using timeout_t = std::chrono::time_point<std::chrono::system_clock>;
 static timeout_t timeout_time(int microseconds) {
@@ -91,8 +96,8 @@ class x11_base_t {
   void x11_acknowledge_wakeup(fd_set *fds) {
     /* Clear data from wake-up pipe */
     if (FD_ISSET(wakeup_pipe[0], fds)) {
-      char buffer[8];
-      read(wakeup_pipe[0], buffer, sizeof(buffer));
+      char buffer;
+      nosig_read(wakeup_pipe[0], &buffer, sizeof(buffer));
     }
   }
 
@@ -102,7 +107,10 @@ class x11_base_t {
     return wakeup_pipe[0] + 1;
   }
 
-  void send_wakeup() { write(wakeup_pipe[1], &wakeup_pipe, 1); }
+  void send_wakeup() {
+    char data;
+    nosig_write(wakeup_pipe[1], &data, 1);
+  }
 
   x11_atom_t get_atom(int atom) { return atoms[atom]; }
   x11_window_t get_window() { return window; }
@@ -357,9 +365,9 @@ class x11_driver_t {
 
 #define x11_working() (x11.is_initialized() && !x11.has_error())
 
-  linked_ptr<std::string>::t get_selection(bool clipboard) {
+  std::shared_ptr<std::string> get_selection(bool clipboard) {
     timeout_t timeout = timeout_time(1000000);
-    linked_ptr<std::string>::t result;
+    std::shared_ptr<std::string> result;
 
     /* NOTE: the clipboard is supposed to be locked when this routine is called. */
 
@@ -379,7 +387,7 @@ class x11_driver_t {
       x11.x11_flush();
       if (clipboard_signal.wait_until(clipboard_mutex_lock, timeout) != std::cv_status::timeout &&
           conversion_succeeded) {
-        result = new std::string(retrieved_data);
+        result.reset(new std::string(retrieved_data));
       }
       action = ACTION_NONE;
     } else {
@@ -388,20 +396,21 @@ class x11_driver_t {
     return result;
   }
 
-  void claim_selection(bool clipboard, std::string *data) {
+  void claim_selection(bool clipboard, std::unique_ptr<std::string> data) {
     timeout_t timeout = timeout_time(1000000);
 
     if (!x11_working()) {
       if (clipboard) {
-        clipboard_data = data;
+        clipboard_data.reset(data.release());
       } else {
-        primary_data = data;
+        primary_data.reset(data.release());
       }
       return;
     }
 
     std::unique_lock<std::mutex> l(clipboard_mutex);
 
+    const bool has_data = data != nullptr;
     if (clipboard) {
       /* If we don't own the selection, reseting is a no-op. */
       if (clipboard_owner_since == X11_CURRENT_TIME && data == nullptr) {
@@ -409,17 +418,17 @@ class x11_driver_t {
       }
 
       action = CLAIM_CLIPBOARD;
-      clipboard_data = data;
+      clipboard_data.reset(data.release());
     } else {
       /* If we don't own the selection, reseting is a no-op. */
       if (primary_owner_since == X11_CURRENT_TIME && data == nullptr) {
         return;
       }
       action = CLAIM_PRIMARY;
-      primary_data = data;
+      primary_data.reset(data.release());
     }
 
-    if (data != nullptr) {
+    if (has_data) {
       x11.x11_change_property(x11.get_window(), X11_ATOM_WM_NAME, X11_ATOM_STRING, 8,
                               X11_PROPERTY_APPEND, nullptr, 0);
     } else {
@@ -514,7 +523,7 @@ class x11_driver_t {
           @return A boolean indicating succes.
   */
   bool send_selection(x11_window_t requestor, x11_atom_t target, x11_atom_t property,
-                      linked_ptr<std::string>::t data, x11_time_t since) {
+                      std::shared_ptr<std::string> data, x11_time_t since) {
     if (target == x11.get_atom(TARGETS)) {
       x11.x11_change_property(requestor, property, x11.get_atom(ATOM), 32, X11_PROPERTY_REPLACE,
                               x11.get_targets_list(), 4);
@@ -749,7 +758,7 @@ class x11_driver_t {
         }
         case X11_SELECTION_REQUEST: {
           x11_selection_event_t reply;
-          linked_ptr<std::string>::t data;
+          std::shared_ptr<std::string> data;
           x11_time_t since;
           x11_selection_request_event_t *request_event =
               reinterpret_cast<x11_selection_request_event_t *>(event);
@@ -814,11 +823,12 @@ class x11_driver_t {
 
   struct incr_send_data_t {
     x11_window_t window;
-    linked_ptr<std::string>::t data;
+    std::shared_ptr<std::string> data;
     x11_atom_t property;
     size_t offset;
 
-    incr_send_data_t(x11_window_t _window, linked_ptr<std::string>::t &_data, x11_atom_t _property)
+    incr_send_data_t(x11_window_t _window, std::shared_ptr<std::string> &_data,
+                     x11_atom_t _property)
         : window(_window), data(_data), property(_property), offset(0) {}
   };
   using incr_send_list_t = std::list<incr_send_data_t>;
@@ -859,18 +869,18 @@ static void release_selections() {
   x11_driver_t::implementation->release_selections();
 }
 
-static linked_ptr<std::string>::t get_selection(bool clipboard) {
+static std::shared_ptr<std::string> get_selection(bool clipboard) {
   if (!x11_driver_t::implementation) {
     return clipboard ? clipboard_data : primary_data;
   }
   return x11_driver_t::implementation->get_selection(clipboard);
 }
 
-static void claim_selection(bool clipboard, std::string *data) {
+static void claim_selection(bool clipboard, std::unique_ptr<std::string> data) {
   if (!x11_driver_t::implementation) {
     return;
   }
-  x11_driver_t::implementation->claim_selection(clipboard, data);
+  x11_driver_t::implementation->claim_selection(clipboard, std::move(data));
 }
 
 static void lock() {
@@ -907,4 +917,4 @@ T3_WIDGET_API extclipboard_interface_t _t3_widget_extclipboard_calls = {EXTCLIPB
                                                                         stop_x11};
 };
 
-};  // namespace
+}  // namespace t3widget
